/* /////////////////////////////////////////////////////////////////////////////////////////////////
Life#IdlePoser v0.1 by MacGruber.
State-based idle animation system with anchoring.

Version 0.1 2020-05-24
	Initial prototype release.

///////////////////////////////////////////////////////////////////////////////////////////////// */

using UnityEngine;
using UnityEngine.Events;
using System;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
	public partial class IdlePoser : MVRScript
	{
		private const int MAX_STATES = 4;

		private Dictionary<string, State> myStates = new Dictionary<string, State>();
		private List<ControlCapture> myControlCaptures = new List<ControlCapture>();
		private State myCurrentState;
		private State myNextState;
		private State myBlendState = new State() {
			myName = "BlendState",
			myWaitDurationMin = 0.0f,
			myWaitDurationMax = 0.0f,
			myTransitionDuration = 0.2f,
			myIsTransition = true
		};

		private float myDuration = 1.0f;
		private float myClock = 0.0f;
		private bool myPaused = false;

		public override void Init()
		{			
			myClock = 0.0f;

			myControlCaptures.Add(new ControlCapture(this, "lHandControl"));
			myControlCaptures.Add(new ControlCapture(this, "rHandControl"));

			InitUI();
		}

		private State CreateState(string name)
		{
			State s = new State() {
				myName = name,
				myWaitDurationMin = 3.0f,
				myWaitDurationMax = 5.0f,
				myTransitionDuration = 0.5f,
				myIsTransition = false
			};
			CaptureState(s);
			myStates[name] = s;
			return s;
		}

		private void SetState(State state)
		{
			myCurrentState = state;
			myNextState = null;
			myDuration = UnityEngine.Random.Range(state.myWaitDurationMin, state.myWaitDurationMax);
		}

		private void SetTransition(List<State> states)
		{
			float duration = 0.0f;
			for (int i=0; i<states.Count; ++i)
				duration += states[i].myTransitionDuration;
			SetTransition(states, duration);
		}

		private void SetTransition(List<State> states, float duration)
		{		
			myNextState = states[states.Count-1];
			myDuration = Mathf.Max(duration, 0.001f);
			for (int i=0; i<myControlCaptures.Count; ++i)
				myControlCaptures[i].SetTransition(states);
		}

		private void SetRandomTransition()
		{
			List<State> states = new List<State>(MAX_STATES);
			states.Add(myCurrentState);
			GatherRandomTransition(states);
			SetTransition(states);
		}

		private void GatherRandomTransition(List<State> states)
		{
			State s = states[states.Count-1];
			List<State> possible = new List<State>(s.myTransitions.Count);
			for (int i=0; i<s.myTransitions.Count; ++i)
			{
				if (!states.Contains(s.myTransitions[i]))
					possible.Add(s.myTransitions[i]);
			}
			if (possible.Count == 0)
				return;

			int idx = UnityEngine.Random.Range(0, possible.Count);
			s = possible[idx];
			states.Add(s);

			if (s.myIsTransition && states.Count < MAX_STATES)
				GatherRandomTransition(states);
		}

		private bool FindTransitionToState(State source, State target, HashSet<State> visited, List<State> transition)
		{
			for (int i=0; i<source.myTransitions.Count; ++i)
			{
				State current = source.myTransitions[i];
				if (current == target)
				{
					transition.Add(target);
					return true;
				}
				else if (current.myIsTransition && transition.Count <= MAX_STATES-2 && !visited.Contains(current))
				{
					transition.Add(current);
					visited.Add(current);
					if (FindTransitionToState(current, target, visited, transition))
						return true;
					else
						transition.RemoveAt(transition.Count-1);
				}
			}
			return false;
		}

		private void BlendTransitionToState(State state)
		{
			CaptureState(myBlendState);
			List<State> transition = new List<State>(MAX_STATES);
			transition.Add(myBlendState);

			HashSet<State> visited = new HashSet<State>();
			if (myCurrentState == null || !FindTransitionToState(myCurrentState, state, visited, transition))
				transition.Add(state);

			myClock = 0;
			SetState(myBlendState);
			SetTransition(transition);
		}

		private void BlendToRandomState(float duration)
		{
			List<State> possible = new List<State>(myStates.Count);
			foreach (var state in myStates)
			{
				if (!state.Value.myIsTransition)
					possible.Add(state.Value);
			}
			int idx = UnityEngine.Random.Range(0, possible.Count);

			CaptureState(myBlendState);
			List<State> states = new List<State>(2);
			states.Add(myBlendState);
			states.Add(possible[idx]);
			SetState(myBlendState);
			SetTransition(states, duration);
		}

		private void CaptureState(State state)
		{
			for (int i=0; i<myControlCaptures.Count; ++i)
				myControlCaptures[i].CaptureEntry(state);
		}

		private void Update()
		{
			if (myCurrentState == null)
				return;

			if (myNextState != null || !myPaused)
				myClock += Time.deltaTime;

			if (myNextState != null)
			{
				float t = Mathf.InverseLerp(0.0f, myDuration, myClock);
				t = Mathf.Clamp01(t);
				//t = SmootherStep(t);
				t = Mathf.SmoothStep(0.0f, 1.0f, t);
				for (int i=0; i<myControlCaptures.Count; ++i)
					myControlCaptures[i].UpdateTransition(t);
			}
			else if (!myPaused)
			{
				for (int i=0; i<myControlCaptures.Count; ++i)
					myControlCaptures[i].UpdateState(myCurrentState);
			}

			if (myClock >= myDuration)
			{
				myClock -= myDuration;
				if (myNextState != null)
					SetState(myNextState);
				else if (!myPaused)
					SetRandomTransition();
			}
		}

		private static float SmootherStep(float t)
		{
			return t * t * t * (t * (t * 6.0f - 15.0f) + 10.0f);
		}

		private void SaveJSON(string path)
		{
			JSONClass jc = new JSONClass();

			JSONClass info = new JSONClass();
			info["Format"] = "MacGruber.Life.IdlePoser";
			info["Version"].AsInt = 1;
			string creatorName = UserPreferences.singleton.creatorName;
			if (string.IsNullOrEmpty(creatorName))
				creatorName = "Unknown";
			info["Author"] = creatorName;
			jc["Info"] = info;

			if (myCurrentState != null)
				jc["InitialState"] = myCurrentState.myName;

			JSONArray cclist = new JSONArray();
			for (int i=0; i<myControlCaptures.Count; ++i)
				cclist.Add("", myControlCaptures[i].myName);
			jc["ControlCaptures"] = cclist;

			JSONArray slist = new JSONArray();
			foreach (var s in myStates)
			{
				State state = s.Value;
				JSONClass st = new JSONClass();
				st["Name"] = state.myName;
				st["WaitDurationMin"].AsFloat = state.myWaitDurationMin;
				st["WaitDurationMax"].AsFloat = state.myWaitDurationMax;
				st["TransitionDuration"].AsFloat = state.myTransitionDuration;
				st["IsTransition"].AsBool = state.myIsTransition;

				JSONArray tlist = new JSONArray();
				for (int i=0; i<state.myTransitions.Count; ++i)
					tlist.Add("", state.myTransitions[i].myName);
				st["Transitions"] = tlist;

				JSONClass celist = new JSONClass();
				foreach (var e in state.myControlEntries)
				{
					ControlEntryAnchored ce = e.Value;
					JSONClass ceclass = new JSONClass();
					ceclass["PX"].AsFloat = ce.myAnchorOffset.myPosition.x;
					ceclass["PY"].AsFloat = ce.myAnchorOffset.myPosition.y;
					ceclass["PZ"].AsFloat = ce.myAnchorOffset.myPosition.z;
					Vector3 rotation = ce.myAnchorOffset.myRotation.eulerAngles;
					ceclass["RX"].AsFloat = rotation.x;
					ceclass["RY"].AsFloat = rotation.y;
					ceclass["RZ"].AsFloat = rotation.z;
					ceclass["AnchorMode"].AsInt = ce.myAnchorMode;
					ceclass["BlendRatio"].AsFloat = ce.myBlendRatio;
					ceclass["DampingTime"].AsFloat = ce.myDampingTime;
					ceclass["AnchorAAtom"] = ce.myAnchorAAtom;
					ceclass["AnchorAControl"] = ce.myAnchorAControl;
					ceclass["AnchorBAtom"] = ce.myAnchorBAtom;
					ceclass["AnchorBControl"] = ce.myAnchorBControl;

					celist[e.Key.myName] = ceclass;
				}
				st["ControlEntries"] = celist;

				slist.Add("", st);
			}
			jc["States"] = slist;

			SuperController.singleton.SaveJSON(jc, path);
		}

		private void LoadJSON(string path)
		{
			JSONClass jc = SuperController.singleton.LoadJSON(path).AsObject;
			if (jc == null)
				return;

			// reset
			myControlCaptures.Clear();
			myStates.Clear();
			myCurrentState = null;
			myNextState = null;
			myBlendState.myControlEntries.Clear();
			myClock = 0.0f;

			// load captures
			JSONArray cclist = jc["ControlCaptures"].AsArray;
			for (int i=0; i<cclist.Count; ++i)
			{
				ControlCapture cc = new ControlCapture(this, cclist[i].Value);
				if (cc.IsValid())
					myControlCaptures.Add(cc);
			}

			// load states
			JSONArray slist = jc["States"].AsArray;
			for (int i=0; i<slist.Count; ++i)
			{
				JSONClass st = slist[i].AsObject;

				State state = new State() {
					myName = st["Name"],
					myWaitDurationMin = st["WaitDurationMin"].AsFloat,
					myWaitDurationMax = st["WaitDurationMax"].AsFloat,
					myTransitionDuration = st["TransitionDuration"].AsFloat,
					myIsTransition = st["IsTransition"].AsBool
				};

				if (myStates.ContainsKey(state.myName))
					continue;
				myStates[state.myName] = state;

				JSONClass celist = st["ControlEntries"].AsObject;
				foreach (string ccname in celist.Keys)
				{
					ControlCapture cc = myControlCaptures.Find(x => x.myName == ccname);
					if (cc == null)
						continue;

					JSONClass ceclass = celist[ccname].AsObject;
					ControlEntryAnchored ce = new ControlEntryAnchored();
					ce.myAnchorOffset.myPosition.x = ceclass["PX"].AsFloat;
					ce.myAnchorOffset.myPosition.y = ceclass["PY"].AsFloat;
					ce.myAnchorOffset.myPosition.z = ceclass["PZ"].AsFloat;
					Vector3 rotation;
					rotation.x = ceclass["RX"].AsFloat;
					rotation.y = ceclass["RY"].AsFloat;
					rotation.z = ceclass["RZ"].AsFloat;
					ce.myAnchorOffset.myRotation.eulerAngles = rotation;
					ce.myAnchorMode = ceclass["AnchorMode"].AsInt;
					ce.myBlendRatio = ceclass["BlendRatio"].AsFloat;
					ce.myDampingTime = ceclass["DampingTime"].AsFloat;
					ce.myAnchorAAtom = ceclass["AnchorAAtom"].Value;
					ce.myAnchorAControl = ceclass["AnchorAControl"].Value;
					ce.myAnchorBAtom = ceclass["AnchorBAtom"].Value;
					ce.myAnchorBControl = ceclass["AnchorBControl"].Value;
					ce.Initialize(this);

					state.myControlEntries.Add(cc, ce);
				}

				for (int j=0; j<myControlCaptures.Count; ++j)
				{
					if (!state.myControlEntries.ContainsKey(myControlCaptures[j]))
						myControlCaptures[j].CaptureEntry(state);
				}
			}

			// load transitions
			for (int i=0; i<slist.Count; ++i)
			{
				JSONClass st = slist[i].AsObject;
				State source;
				if (!myStates.TryGetValue(st["Name"], out source))
					continue;

				JSONArray tlist = st["Transitions"].AsArray;
				for (int j=0; j<tlist.Count; ++j)
				{
					State target;
					if (myStates.TryGetValue(tlist[j].Value, out target))
						source.myTransitions.Add(target);
				}
			}

			// blend to initial state
			if (jc.HasKey("InitialState"))
			{
				State initial;
				if (myStates.TryGetValue(jc["InitialState"].Value, out initial))
					myCurrentState = initial;
			}
		}

		// =======================================================================================


		private Atom GetSelfOrAtomById(string atomName)
		{
			return atomName == "[Self]" ? containingAtom : GetAtomById(atomName);
		}


		// =======================================================================================


		private class State
		{
			public string myName;
			public float myWaitDurationMin;
			public float myWaitDurationMax;
			public float myTransitionDuration;
			public bool myIsTransition;
			public Dictionary<ControlCapture, ControlEntryAnchored> myControlEntries = new Dictionary<ControlCapture, ControlEntryAnchored>();
			public List<State> myTransitions = new List<State>();
		}

		private class ControlCapture
		{
			public string myName;
			private IdlePoser myPlugin;
			private Transform myTransform;
			private ControlEntryAnchored[] myTransition = new ControlEntryAnchored[MAX_STATES];
			private int myEntryCount = 0;

			private static ControlEntry[][] ourTempEntries = new ControlEntry[][] {
				new ControlEntry[3],
				new ControlEntry[2],
				new ControlEntry[1]
			};

			public ControlCapture(IdlePoser plugin, string control)
			{
				myPlugin = plugin;
				myName = control;
				FreeControllerV3 controller = plugin.containingAtom.GetStorableByID(control) as FreeControllerV3;
				if (controller != null)
					myTransform = controller.transform;
			}

			public void CaptureEntry(State state)
			{
				ControlEntryAnchored entry;
				if (!state.myControlEntries.TryGetValue(this, out entry))
				{
					entry = new ControlEntryAnchored();
					entry.Initialize(myPlugin);
					state.myControlEntries[this] = entry;
				}
				entry.Capture(myTransform.position, myTransform.rotation);
			}

			public void SetTransition(List<State> states)
			{
				myEntryCount = states.Count;
				for (int i=0; i<myEntryCount; ++i)
				{
					if (!states[i].myControlEntries.TryGetValue(this, out myTransition[i]))
					{
						CaptureEntry(states[i]);
						myTransition[i] = states[i].myControlEntries[this];
					}
				}
			}

			public void UpdateTransition(float t)
			{
				for (int i=0; i<myEntryCount; ++i)
					myTransition[i].Update();

				if (myEntryCount >= 2)
				{
					// evaluating Bézier curve
					ControlEntry[] curr = ourTempEntries[0];
					for (int i=0; i<myEntryCount-1; ++i)
					{
						curr[i].Lerp(myTransition[i].myEntry, myTransition[i+1].myEntry, t);
					}
					for (int level=1; level<myEntryCount-1; ++level)
					{
						ControlEntry[] prev = curr;
						curr = ourTempEntries[level];
						for (int i=0; i<myEntryCount-level-1; ++i)
							curr[i].Lerp(prev[i], prev[i+1], t);
					}
					curr[0].SetTransform(myTransform);
				}
				else if (myEntryCount == 1)
				{
					myTransition[0].myEntry.SetTransform(myTransform);
				}
			}

			public void UpdateState(State state)
			{
				ControlEntryAnchored entry;
				if (state.myControlEntries.TryGetValue(this, out entry))
				{
					entry.Update();
					entry.myEntry.SetTransform(myTransform);
				}
			}

			public bool IsValid()
			{
				return myTransform != null;
			}
		}

		private struct ControlEntry
		{
			public Quaternion myRotation;
			public Vector3 myPosition;

			public void Lerp(ControlEntry a, ControlEntry b, float t)
			{
				myRotation = Quaternion.SlerpUnclamped(a.myRotation, b.myRotation, t);
				myPosition = Vector3.LerpUnclamped(a.myPosition, b.myPosition, t);
			}

			public void SetTransform(Transform t)
			{
				t.position = myPosition;
				t.rotation = myRotation;
			}
		}

		private class ControlEntryAnchored
		{
			public const int ANCHORMODE_WORLD = 0;
			public const int ANCHORMODE_SINGLE = 1;
			public const int ANCHORMODE_BLEND = 2;

			public ControlEntry myEntry;
			public ControlEntry myAnchorOffset;
			public Transform myAnchorATransform;
			public Transform myAnchorBTransform;
			public int myAnchorMode = 1;
			public float myBlendRatio = 0.5f;
			public float myDampingTime = 0.2f;

			public string myAnchorAAtom = "[Self]";
			public string myAnchorBAtom;
			public string myAnchorAControl = "control";
			public string myAnchorBControl;

			public void Initialize(IdlePoser plugin)
			{
				GetTransforms(plugin);
				UpdateInstant();
			}

			public void AdjustAnchor(IdlePoser plugin)
			{
				UpdateInstant();
				GetTransforms(plugin);
				Capture(myEntry.myPosition, myEntry.myRotation);
			}

			private void GetTransforms(IdlePoser plugin)
			{
				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myAnchorATransform = null;
					myAnchorBTransform = null;
				}
				else
				{
					myAnchorATransform = GetTransform(plugin, myAnchorAAtom, myAnchorAControl);
					if (myAnchorMode == ANCHORMODE_BLEND)
						myAnchorBTransform = GetTransform(plugin, myAnchorBAtom, myAnchorBControl);
					else
						myAnchorBTransform = null;
				}
			}

			private Transform GetTransform(IdlePoser plugin, string atomName, string controlName)
			{
				Atom atom = plugin.GetSelfOrAtomById(atomName);
				return atom?.GetStorableByID(controlName)?.transform;
			}

			public void UpdateInstant()
			{
				float dampingTime = myDampingTime;
				myDampingTime = 0.0f;
				Update();
				myDampingTime = dampingTime;
			}

			public void Update()
			{
				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myEntry = myAnchorOffset;
				}
				else
				{
					ControlEntry anchor;
					if (myAnchorMode == ANCHORMODE_SINGLE)
					{
						if (myAnchorATransform == null)
							return;
						anchor.myPosition = myAnchorATransform.position;
						anchor.myRotation = myAnchorATransform.rotation;
					}
					else
					{
						if (myAnchorATransform == null || myAnchorBTransform == null)
							return;
						anchor.myPosition = Vector3.LerpUnclamped(myAnchorATransform.position, myAnchorBTransform.position, myBlendRatio);
						anchor.myRotation = Quaternion.SlerpUnclamped(myAnchorATransform.rotation, myAnchorBTransform.rotation, myBlendRatio);
					}
					anchor.myPosition = anchor.myPosition + anchor.myRotation * myAnchorOffset.myPosition;
					anchor.myRotation = anchor.myRotation * myAnchorOffset.myRotation;

					if (myDampingTime >= 0.001f)
					{
						float t = Mathf.Clamp01(Time.deltaTime / myDampingTime);
						myEntry.Lerp(myEntry, anchor, t);
					}
					else
					{
						myEntry = anchor;
					}
				}
			}

			public void Capture(Vector3 position, Quaternion rotation)
			{
				myEntry.myPosition = position;
				myEntry.myRotation = rotation;

				if (myAnchorMode == ANCHORMODE_WORLD)
				{
					myAnchorOffset.myPosition = position;
					myAnchorOffset.myRotation = rotation;
				}
				else
				{
					ControlEntry root;
					if (myAnchorMode == ANCHORMODE_SINGLE)
					{
						if (myAnchorATransform == null)
							return;
						root.myPosition = myAnchorATransform.position;
						root.myRotation = myAnchorATransform.rotation;
					}
					else
					{
						if (myAnchorATransform == null || myAnchorBTransform == null)
							return;
						root.myPosition = Vector3.LerpUnclamped(myAnchorATransform.position, myAnchorBTransform.position, myBlendRatio);
						root.myRotation = Quaternion.SlerpUnclamped(myAnchorATransform.rotation, myAnchorBTransform.rotation, myBlendRatio);
					}

					myAnchorOffset.myPosition = Quaternion.Inverse(root.myRotation) * (position - root.myPosition);
					myAnchorOffset.myRotation = Quaternion.Inverse(root.myRotation) * rotation;
				}
			}
		}
	}
}